# grid gauge in 3d view (only ortho and parallel to one of axis)

bl_info = {
		"name": "Grid Gauge",
		"blender": (2, 7, 3),
		"version": (0, 0, 3),
		"category": "3D View",
};

import bpy;
import bgl;
import blf;
import math;
from mathutils import *;

# float range generator
def frange(x, y, jump):
	x = float(x);
	y = float(y);
	while x < y:
		yield x
		x += jump

def draw_callback(self, context):
	def rectangle_check(x, y, width, height):
		return x > 0 and y > 0 and x < width and y < height;

	def epscheck(x, chk, eps):
		return (chk - eps) < x and (chk + eps) > x;

	def draw(vec, unit_vec, x_axis, snap, c):
		x = int(mid_x + vec[0] * mid_x);
		y = int(mid_y + vec[1] * mid_y);

		# prettier formatting
		if abs(c) >= 1000:
			c /= 1000;
			strc = "%gk" % c;
		else:
			strc = "%.3g" % c;

		(fw, fh) = blf.dimensions(0, strc);

		if epscheck(c, 0.0, 1e-6):
			(fw, fh) = blf.dimensions(0, "0");
			if snap > 0:
				blf.position(0, width - fw, y, 0);
				blf.draw(0, "0");
				blf.position(0, x, height - fh, 0);
				blf.draw(0, "0");
			else:
				blf.position(0, fw / 2, y, 0);
				blf.draw(0, "0");
				blf.position(0, x, fh / 2, 0);
				blf.draw(0, "0");
			return;

		if snap > 0:
			if abs(unit_vec[0]) < abs(unit_vec[1]):
				x = width - fw;
			else:
				y = height - fh;
		else:
			if abs(unit_vec[0]) < abs(unit_vec[1]):
				x = fw / 2;
			else:
				y = fh / 2;

		blf.position(0, x, y, 0);
		blf.draw(0, strc);

	def draw_point(unit_vector, c):
		axis = view_matrix * (unit_vector * c);
		draw(axis, view_matrix * (unit_vector * 1e9), 1, 1, c);
		axis = view_matrix * (unit_vector * c);
		draw(axis, view_matrix * (unit_vector * 1e9), 1, -1, c);


	def draw_axis(unit_vector):
		rangestart = min(leftdown_axis[:]) // scale * scale - scale;
		rangeend = max(rightup_axis[:]) // scale * scale + scale;
		for i in frange(rangestart, rangeend, scale):
			draw_point(unit_vector, i);

	bgl.glColor3f(1, 1, 1);

	# drawing [visible] region
	region = context.region;
	mid_x = region.width / 2;
	mid_y = region.height / 2;
	width = region.width;
	height = region.height;

	# don't draw if rotation isn't aligned to axis
	if context.region_data.view_perspective != "ORTHO":
		return;
	qs = [
		Quaternion((0.5, 0.5, 0.5, 0.5)),
		Quaternion((1.0, 0.0, 0.0, 0.0)),
		Quaternion((0.7071, 0.7071, 0.0, 0.0))
	];
	vq = context.region_data.view_rotation;
	m = 0;
	for q in qs:
		lm = 0;
		for i in range(len(q[:])):
			if epscheck(vq[:][i], q[:][i], 0.0001):
				lm += 1;
		if lm == len(q[:]):
			m += 1;
	if m == 0:
		return;

	view_matrix = context.region_data.perspective_matrix;

	iv = view_matrix.copy();
	iv.invert();
	rightup_axis = iv * Vector((1, 1, 0));
	leftdown_axis = iv * Vector((-1, -1, 0));

	#lines = context.space_data.grid_lines;
	scale = context.space_data.grid_scale;

	eulers = view_matrix.to_euler();

	eulers[0] %= math.pi;
	eulers[1] %= math.pi;
	eulers[2] %= math.pi;

	unit_vectors = [];

	axis = view_matrix * Vector((1e9, 0.0, 0.0));
	if epscheck(axis[2], 0.0, 1):
		unit_vectors.append(Vector((1, 0, 0)));

	axis = view_matrix * Vector((0.0, 1e9, 0.0));
	if epscheck(axis[2], 0.0, 1):
		unit_vectors.append(Vector((0, 1, 0)));

	axis = view_matrix * Vector((0.0, 0.0, 1e9));
	if epscheck(axis[2], 0.0, 1):
		unit_vectors.append(Vector((0, 0, 1)));

	scales = [];
	for uv in unit_vectors:
		a0 = view_matrix * Vector((0, 0, 0));
		a1 = view_matrix * uv;
		s = scale;
		if (a0 - a1).length < 0.035:
			s = scale * (10 ** (1 + int(math.log10(int(0.035 / (a0 - a1).length)))));
			if s > (10 ** context.scene.display_grid_gauge_max_scale):
				s = 10 ** context.scene.display_grid_gauge_max_scale;
			#print("%f %f %f" % (s, scale, (a0 - a1).length));
		elif (a0 - a1).length > 0.35:
			s = scale / (10 ** (1 + int(math.log10(int((a0 - a1).length / 0.35)))));
			if s < (0.1 ** context.scene.display_grid_gauge_min_scale):
				s = 0.1 ** context.scene.display_grid_gauge_min_scale;
			#print("%f %f %f" % (s, scale, (a0 - a1).length));
		scales.append(s);

	if len(scales):
		scale = max(scales);
		for uv in unit_vectors:
			draw_axis(uv);

class visualiser(bpy.types.Operator):
	bl_idname = "view3d.grid_gauge_labels";
	bl_label = "grid gauge visualiser";

	draw_callback_id = None;
	inv = bpy.props.BoolProperty(default=False);

	def execute(self, context):
		b = context.scene.display_grid_gauge_active;
		if self.inv:
			b = not b;

		if(b):
			if visualiser.draw_callback_id:
				bpy.types.SpaceView3D.draw_handler_remove(visualiser.draw_callback_id, "WINDOW");
				visualiser.draw_callback_id = None;
		else:
			visualiser.draw_callback_id = bpy.types.SpaceView3D.draw_handler_add(draw_callback,
					(self, context), "WINDOW", "POST_PIXEL");

		if not self.inv:
			global skip_display_grid_gauge_active_update;
			skip_display_grid_gauge_active_update = True;
			context.scene.display_grid_gauge_active = not context.scene.display_grid_gauge_active;
			skip_display_grid_gauge_active_update = False;

		if context.area:
			context.area.tag_redraw();

		return {"FINISHED"};


def menu_func(self, context):
	self.layout.separator();
	self.layout.prop(context.scene, "display_grid_gauge_active", toggle=True);
	self.layout.prop(context.scene, "display_grid_gauge_min_scale");
	self.layout.prop(context.scene, "display_grid_gauge_max_scale");

"""
class grid_gauge_prefs(bpy.types.AddonPreferences):
	bl_idname = __name__;
	activate_on_startup = bpy.props.BoolProperty(name="Activate on startup");

	def draw(self, context):
		l = self.layout;
		l.prop(self, "activate_on_startup");

		l.label("Toggle: cmd+g");
"""

skip_display_grid_gauge_active_update = False;
def update_grid_gauge_bool(self, context):
	global skip_display_grid_gauge_active_update;
	if(not skip_display_grid_gauge_active_update):
		bpy.ops.view3d.grid_gauge_labels(inv=True);


addon_keymaps = [];

def register():
	bpy.types.Scene.display_grid_gauge_min_scale = bpy.props.IntProperty(
			name="Lower Digits",
			min=0,
			default=1);
	bpy.types.Scene.display_grid_gauge_max_scale = bpy.props.IntProperty(
			name="Higher Digits",
			min=0,
			default=3);
	bpy.types.Scene.display_grid_gauge_active = bpy.props.BoolProperty(
			name="Grid Labels",
			update=update_grid_gauge_bool);

	#bpy.utils.register_class(grid_gauge_prefs);
	bpy.utils.register_class(visualiser);

	bpy.types.VIEW3D_PT_view3d_display.append(menu_func);

	# autostart
	#bpy.app.handlers.scene_update_pre.append(init);

	# keys
	global addon_keymaps;
	wm = bpy.context.window_manager;
	km = wm.keyconfigs.addon.keymaps.new(name="3D View", space_type="VIEW_3D");
	kmi = km.keymap_items.new(visualiser.bl_idname, 'G', 'RELEASE', oskey=True);
	addon_keymaps.append((km, kmi));

def unregister():
	global addon_keymaps;
	for km, kmi in addon_keymaps:
		km.keymap_items.remove(kmi);
	addon_keymans.clear();

	#bpy.utils.unregister_class(grid_gauge_prefs);
	bpy.utils.unregister_class(visualiser);
	bpy.types.VIEW3D_PT_view3d_display.remove(menu_func);

	if visualiser.draw_callback_id:
		bpy.types.SpaceView3D.draw_handler_remove(visualiser.draw_callback_id, "WINDOW");

"""
# autostart
@bpy.app.handlers.persistent
def init(arg):
	if init in bpy.app.handlers.scene_update_pre:
		bpy.app.handlers.scene_update_pre.remove(init);

	if bpy.context.user_preferences.addons[__name__].preferences.activate_on_startup:
		bpy.ops.view3d.grid_gauge_labels();
"""

if __name__ == "__main__":
	register();
